@@ -2,32 +2,18 @@
 Copyright 2007-2011, 2016q Free Software Foundation, Inc.
 This file is part of GNU Radio
 
-GNU Radio Companion is free software; you can redistribute it and/or
-modify it under the terms of the GNU General Public License
-as published by the Free Software Foundation; either version 2
-of the License, or (at your option) any later version.
+SPDX-License-Identifier: GPL-2.0-or-later
 
-GNU Radio Companion is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program; if not, write to the Free Software
-Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 """
 
-from __future__ import absolute_import
 
 import ast
 import functools
 import random
-from distutils.spawn import find_executable
+from shutil import which as find_executable
 from itertools import count
 
-import six
 from gi.repository import GLib, Gtk
-from six.moves import filter
 
 from . import colors
 from .drawable import Drawable
@@ -49,7 +35,7 @@ class _ContextMenu(object):
 
         # In GTK 3.22 Menu.popup was deprecated, we want to popup at the
         # pointer, so use that new function instead if we can.
-        if Gtk.check_version(3,22,0) is None:
+        if Gtk.check_version(3, 22, 0) is None:
             self.popup = self._menu.popup_at_pointer
 
     def popup(self, event):
@@ -139,9 +125,18 @@ class FlowGraph(CoreFlowgraph, Drawable):
             editor.open_editor()
         except Exception as e:
             # Problem launching the editor. Need to select a new editor.
-            Messages.send('>>> Error opening an external editor. Please select a different editor.\n')
+            Messages.send(
+                '>>> Error opening an external editor. Please select a different editor.\n')
             # Reset the editor to force the user to select a new one.
             self.parent_platform.config.editor = ''
+            self.remove_external_editor(target=target)
+
+    def remove_external_editor(self, target=None, param=None):
+        if target is None:
+            target = (param.parent_block.name, param.key)
+        if target in self._external_updaters:
+            self._external_updaters[target].stop()
+            del self._external_updaters[target]
 
     def handle_external_editor_change(self, new_value, target):
         try:
@@ -149,8 +144,7 @@ class FlowGraph(CoreFlowgraph, Drawable):
             self.get_block(block_id).params[param_key].set_value(new_value)
 
         except (IndexError, ValueError):  # block no longer exists
-            self._external_updaters[target].stop()
-            del self._external_updaters[target]
+            self.remove_external_editor(target=target)
             return
         Actions.EXTERNAL_UPDATE()
 
@@ -167,10 +161,13 @@ class FlowGraph(CoreFlowgraph, Drawable):
         # calculate the position coordinate
         h_adj = scroll_pane.get_hadjustment()
         v_adj = scroll_pane.get_vadjustment()
-        if coor is None: coor = (
-            int(random.uniform(.25, .75)*h_adj.get_page_size() + h_adj.get_value()),
-            int(random.uniform(.25, .75)*v_adj.get_page_size() + v_adj.get_value()),
-        )
+        if coor is None:
+            coor = (
+                int(random.uniform(.25, .75) *
+                    h_adj.get_page_size() + h_adj.get_value()),
+                int(random.uniform(.25, .75) *
+                    v_adj.get_page_size() + v_adj.get_value()),
+            )
         # get the new block
         block = self.new_block(key)
         block.coordinate = coor
@@ -232,17 +229,17 @@ class FlowGraph(CoreFlowgraph, Drawable):
         Returns:
             the clipboard
         """
-        #get selected blocks
+        # get selected blocks
         blocks = list(self.selected_blocks())
         if not blocks:
             return None
-        #calc x and y min
+        # calc x and y min
         x_min, y_min = blocks[0].coordinate
         for block in blocks:
             x, y = block.coordinate
             x_min = min(x, x_min)
             y_min = min(y, y_min)
-        #get connections between selected blocks
+        # get connections between selected blocks
         connections = list(filter(
             lambda c: c.source_block in blocks and c.sink_block in blocks,
             self.connections,
@@ -261,10 +258,7 @@ class FlowGraph(CoreFlowgraph, Drawable):
         Args:
             clipboard: the nested data of blocks, connections
         """
-        # todo: rewrite this...
-        selected = set()
         (x_min, y_min), blocks_n, connections_n = clipboard
-        old_id2block = dict()
         # recalc the position
         scroll_pane = self.drawing_area.get_parent().get_parent()
         h_adj = scroll_pane.get_hadjustment()
@@ -276,6 +270,7 @@ class FlowGraph(CoreFlowgraph, Drawable):
             x_off, y_off = 0, 0
 
         # create blocks
+        pasted_blocks = {}
         for block_n in blocks_n:
             block_key = block_n.get('id')
             if block_key == 'options':
@@ -284,30 +279,35 @@ class FlowGraph(CoreFlowgraph, Drawable):
             block_name = block_n.get('name')
             # Verify whether a block with this name exists before adding it
             if block_name in (blk.name for blk in self.blocks):
-                block_name = self._get_unique_id(block_name)
-                block_n['name'] = block_name
+                block_n = block_n.copy()
+                block_n['name'] = self._get_unique_id(block_name)
 
             block = self.new_block(block_key)
             if not block:
                 continue  # unknown block was pasted (e.g. dummy block)
 
-            selected.add(block)
             block.import_data(**block_n)
-            old_id2block[block.params['id'].value] = block
-            # move block to offset coordinate
-            block.move((x_off, y_off))
+            pasted_blocks[block_name] = block  # that is before any rename
 
-            #TODO: prevent block from being pasted directly on top of another block
+            block.move((x_off, y_off))
+            while any(Utils.align_to_grid(block.coordinate) == Utils.align_to_grid(other.coordinate)
+                      for other in self.blocks if other is not block):
+                block.move((Constants.CANVAS_GRID_SIZE,
+                           Constants.CANVAS_GRID_SIZE))
+                # shift all following blocks
+                x_off += Constants.CANVAS_GRID_SIZE
+                y_off += Constants.CANVAS_GRID_SIZE
+
+        self.selected_elements = set(pasted_blocks.values())
 
         # update before creating connections
         self.update()
         # create connections
-        for connection_n in connections_n:
-            source = old_id2block[connection_n[0]].get_source(connection_n[1])
-            sink = old_id2block[connection_n[2]].get_sink(connection_n[3])
+        for src_block, src_port, dst_block, dst_port in connections_n:
+            source = pasted_blocks[src_block].get_source(src_port)
+            sink = pasted_blocks[dst_block].get_sink(dst_port)
             connection = self.connect(source, sink)
-            selected.add(connection)
-        self.selected_elements = selected
+            self.selected_elements.add(connection)
 
     ###########################################################################
     # Modify Selected
@@ -322,7 +322,7 @@ class FlowGraph(CoreFlowgraph, Drawable):
         Returns:
             true for change
         """
-        return any(sb.type_controller_modify(direction) for sb in self.selected_blocks())
+        return any([sb.type_controller_modify(direction) for sb in self.selected_blocks()])
 
     def port_controller_modify_selected(self, direction):
         """
@@ -334,7 +334,7 @@ class FlowGraph(CoreFlowgraph, Drawable):
         Returns:
             true for changed
         """
-        return any(sb.port_controller_modify(direction) for sb in self.selected_blocks())
+        return any([sb.port_controller_modify(direction) for sb in self.selected_blocks()])
 
     def change_state_selected(self, new_state):
         """
@@ -359,7 +359,23 @@ class FlowGraph(CoreFlowgraph, Drawable):
         Args:
             delta_coordinate: the change in coordinates
         """
-        for selected_block in self.selected_blocks():
+
+        # Determine selected blocks top left coordinate
+        blocks = list(self.selected_blocks())
+        if not blocks:
+            return
+
+        min_x, min_y = self.selected_block.coordinate
+        for selected_block in blocks:
+            x, y = selected_block.coordinate
+            min_x, min_y = min(min_x, x), min(min_y, y)
+
+        # Sanitize delta_coordinate so that blocks don't move to negative coordinate
+        delta_coordinate = max(
+            delta_coordinate[0], -min_x), max(delta_coordinate[1], -min_y)
+
+        # Move selected blocks
+        for selected_block in blocks:
             selected_block.move(delta_coordinate)
             self.element_moved = True
 
@@ -385,15 +401,15 @@ class FlowGraph(CoreFlowgraph, Drawable):
             x += selected_block.width
             y += selected_block.height
             max_x, max_y = max(max_x, x), max(max_y, y)
-        ctr_x, ctr_y = (max_x + min_x)/2, (max_y + min_y)/2
+        ctr_x, ctr_y = (max_x + min_x) / 2, (max_y + min_y) / 2
 
         # align the blocks as requested
         transform = {
             Actions.BLOCK_VALIGN_TOP: lambda x, y, w, h: (x, min_y),
-            Actions.BLOCK_VALIGN_MIDDLE: lambda x, y, w, h: (x, ctr_y - h/2),
+            Actions.BLOCK_VALIGN_MIDDLE: lambda x, y, w, h: (x, ctr_y - h / 2),
             Actions.BLOCK_VALIGN_BOTTOM: lambda x, y, w, h: (x, max_y - h),
             Actions.BLOCK_HALIGN_LEFT: lambda x, y, w, h: (min_x, y),
-            Actions.BLOCK_HALIGN_CENTER: lambda x, y, w, h: (ctr_x-w/2, y),
+            Actions.BLOCK_HALIGN_CENTER: lambda x, y, w, h: (ctr_x - w / 2, y),
             Actions.BLOCK_HALIGN_RIGHT: lambda x, y, w, h: (max_x - w, y),
         }.get(calling_action, lambda *args: args)
 
@@ -416,21 +432,22 @@ class FlowGraph(CoreFlowgraph, Drawable):
         """
         if not any(self.selected_blocks()):
             return False
-        #initialize min and max coordinates
+        # initialize min and max coordinates
         min_x, min_y = max_x, max_y = self.selected_block.coordinate
         # rotate each selected block, and find min/max coordinate
         for selected_block in self.selected_blocks():
             selected_block.rotate(rotation)
-            #update the min/max coordinate
+            # update the min/max coordinate
             x, y = selected_block.coordinate
             min_x, min_y = min(min_x, x), min(min_y, y)
             max_x, max_y = max(max_x, x), max(max_y, y)
-        #calculate center point of selected blocks
-        ctr_x, ctr_y = (max_x + min_x)/2, (max_y + min_y)/2
-        #rotate the blocks around the center point
+        # calculate center point of selected blocks
+        ctr_x, ctr_y = (max_x + min_x) / 2, (max_y + min_y) / 2
+        # rotate the blocks around the center point
         for selected_block in self.selected_blocks():
             x, y = selected_block.coordinate
-            x, y = Utils.get_rotated_coordinate((x - ctr_x, y - ctr_y), rotation)
+            x, y = Utils.get_rotated_coordinate(
+                (x - ctr_x, y - ctr_y), rotation)
             selected_block.coordinate = (x + ctr_x, y + ctr_y)
         return True
 
@@ -493,10 +510,10 @@ class FlowGraph(CoreFlowgraph, Drawable):
             element.create_labels(cr)
 
     def create_shapes(self):
-        #TODO - this is a workaround for bus ports not having a proper coordinate
+        # TODO - this is a workaround for bus ports not having a proper coordinate
         # until the shape is drawn.  The workaround is to draw blocks before connections
 
-        for element in filter(lambda x: x.is_block, self._elements_to_draw) :
+        for element in filter(lambda x: x.is_block, self._elements_to_draw):
             element.create_shapes()
 
         for element in filter(lambda x: not x.is_block, self._elements_to_draw):
@@ -505,13 +522,18 @@ class FlowGraph(CoreFlowgraph, Drawable):
     def _drawables(self):
         # todo: cache that
         show_comments = Actions.TOGGLE_SHOW_BLOCK_COMMENTS.get_active()
+        hide_disabled_blocks = Actions.TOGGLE_HIDE_DISABLED_BLOCKS.get_active()
         for element in self._elements_to_draw:
             if element.is_block and show_comments and element.enabled:
                 yield element.draw_comment
         if self._new_connection is not None:
             yield self._new_connection.draw
         for element in self._elements_to_draw:
-            yield element.draw
+            if element not in self.selected_elements:
+                yield element.draw
+        for element in self.selected_elements:
+            if element.enabled or not hide_disabled_blocks:
+                yield element.draw
 
     def draw(self, cr):
         """Draw blocks connections comment and select rectangle"""
@@ -574,7 +596,8 @@ class FlowGraph(CoreFlowgraph, Drawable):
 
         else:  # called from a mouse release
             if not self.element_moved and (not self.selected_elements or self.drawing_area.ctrl_mask) and not self._new_connection:
-                selected_elements = self.what_is_selected(self.coordinate, self.press_coor)
+                selected_elements = self.what_is_selected(
+                    self.coordinate, self.press_coor)
 
         # this selection and the last were ports, try to connect them
         if self.make_connection():
@@ -627,7 +650,8 @@ class FlowGraph(CoreFlowgraph, Drawable):
 
         if selected_port and selected_port.is_source:
             selected.remove(selected_port.parent_block)
-            self._new_connection = DummyConnection(selected_port, coordinate=coor)
+            self._new_connection = DummyConnection(
+                selected_port, coordinate=coor)
             self.drawing_area.queue_draw()
         # update selected ports
         if selected_port is not self._new_selected_port:
@@ -653,7 +677,7 @@ class FlowGraph(CoreFlowgraph, Drawable):
         Returns:
             sub set of blocks in this flow graph
         """
-        return (e for e in self.selected_elements if e.is_block)
+        return (e for e in self.selected_elements.copy() if e.is_block)
 
     @property
     def selected_block(self):
@@ -755,8 +779,6 @@ class FlowGraph(CoreFlowgraph, Drawable):
 
     def _handle_mouse_motion_move(self, coordinate):
         # only continue if mouse-over stuff is enabled (just the auto-hide port label stuff for now)
-        if not Actions.TOGGLE_AUTO_HIDE_PORT_LABELS.get_active():
-            return
         redraw = False
         for element in self._elements_to_draw:
             over_element = element.what_is_selected(coordinate)
@@ -772,6 +794,8 @@ class FlowGraph(CoreFlowgraph, Drawable):
             if self.element_under_mouse:
                 redraw |= self.element_under_mouse.mouse_out() or False
                 self.element_under_mouse = None
+        if not Actions.TOGGLE_AUTO_HIDE_PORT_LABELS.get_active():
+            return
         if redraw:
             # self.create_labels()
             self.create_shapes()
@@ -795,18 +819,33 @@ class FlowGraph(CoreFlowgraph, Drawable):
         x, y = coordinate
         if not self.drawing_area.ctrl_mask:
             X, Y = self.coordinate
-            dX, dY = int(x - X), int(y - Y)
-            active = Actions.TOGGLE_SNAP_TO_GRID.get_active() or self.drawing_area.mod1_mask
-            if not active or abs(dX) >= Constants.CANVAS_GRID_SIZE or abs(dY) >= Constants.CANVAS_GRID_SIZE:
+            dX, dY = x - X, y - Y
+
+            if Actions.TOGGLE_SNAP_TO_GRID.get_active() or self.drawing_area.mod1_mask:
+                dX, dY = int(round(dX / Constants.CANVAS_GRID_SIZE)
+                             ), int(round(dY / Constants.CANVAS_GRID_SIZE))
+                dX, dY = dX * Constants.CANVAS_GRID_SIZE, dY * Constants.CANVAS_GRID_SIZE
+            else:
+                dX, dY = int(round(dX)), int(round(dY))
+
+            if dX != 0 or dY != 0:
                 self.move_selected((dX, dY))
-                self.coordinate = (x, y)
+                self.coordinate = (X + dX, Y + dY)
                 redraw = True
         return redraw
 
     def get_extents(self):
-        extent = 100000, 100000, 0, 0
-        for element in self._elements_to_draw:
-            extent = (min_or_max(xy, e_xy) for min_or_max, xy, e_xy in zip(
-                (min, min, max, max), extent, element.get_extents()
-            ))
+        show_comments = Actions.TOGGLE_SHOW_BLOCK_COMMENTS.get_active()
+
+        def sub_extents():
+            for element in self._elements_to_draw:
+                yield element.get_extents()
+                if element.is_block and show_comments and element.enabled:
+                    yield element.get_extents_comment()
+
+        extent = 10000000, 10000000, 0, 0
+        cmps = (min, min, max, max)
+        for sub_extent in sub_extents():
+            extent = [cmp(xy, e_xy)
+                      for cmp, xy, e_xy in zip(cmps, extent, sub_extent)]
         return tuple(extent)
